/*
	Copyright (C) 2009-2015 Mark Tyler

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program in the file COPYING.
*/

#ifndef MTIMG_H_
#define MTIMG_H_

#include <mtkit.h>



typedef struct mtFont		mtFont;
typedef struct mtImage		mtImage;
typedef struct mtPolygon	mtPolygon;



#define MTIMG_INT_2_R(A)	(((A) >> 16) & 0xFF)
#define MTIMG_INT_2_G(A)	(((A) >> 8) & 0xFF)
#define MTIMG_INT_2_B(A)	((A) & 0xFF)
#define MTIMG_RGB_2_INT(R,G,B)	(((R) << 16) + ((G) << 8) + (B))

#define MTIMG_MAX_WIDTH		30000
#define MTIMG_MAX_HEIGHT	30000



enum
{
	MTIMG_FILE_TYPE_NONE,

	MTIMG_FILE_TYPE_PNG,

	MTIMG_FILE_TYPE_TOTAL
};

enum	// mtFont->type
{
	MTIMG_FONT_NONE		= 0,

	MTIMG_FONT_FIXED,
	MTIMG_FONT_FT_FIXED,
	MTIMG_FONT_FT_SCALABLE,

	MTIMG_FONT_TOTAL
};

enum
{
	MTIMG_POLYGON_FILL,
	MTIMG_POLYGON_OUTLINE
};



// Functions return: 0 = success, NULL = fail; unless otherwise stated.

int mtimg_init (			// Call first before anything else
	void
	);

int mtimg_quit (			// Call once finished with the library
	void
	);

mtPolygon * mtimg_polygon_new (
	void
	);

int mtimg_polygon_add_point (
	mtPolygon	* polygon,
	int		x,
	int		y
	);

int mtimg_polygon_clear (		// Empty a polygon
	mtPolygon	* polygon
	);

int mtimg_polygon_destroy (
	mtPolygon	* polygon
	);

int mtimg_polygon (			// Draw polygon on image
	mtImage		* image,
	int		style,		// MTIMG_POLYGON_*
	mtPolygon	* polygon,
	int		r,
	int		g,
	int		b,
	int		a
	);

mtImage * mtimg_image_new_rgb (
	int		width,		// 1 or more (Pixels)
	int		height		// 1 or more (Pixels)
	);

mtImage * mtimg_image_load (
	char	const	* filename
	);

int mtimg_image_save (
	mtImage		* image,
	char	const	* filename,
	int		filetype,	// MTIMG_FILE_TYPE_*
	int		fileinfo	// e.g. PNG compression level
	);

mtImage * mtimg_image_duplicate (
	mtImage		* image
	);

int mtimg_image_destroy (
	mtImage		* image
	);

int mtimg_image_get_width (
	mtImage		* image
	);
	// 0 = No width, else the image width

int mtimg_image_get_height (
	mtImage		* image
	);
	// 0 = No height, else the image height

unsigned char * mtimg_image_get_rgb (
	mtImage		* image
	);
	// NULL = No RGB, else the image RGB memory

unsigned char * mtimg_image_get_alpha (
	mtImage		* image
	);
	// NULL = No alpha, else the image alpha memory

int mtimg_image_set_data (
	mtImage		* image,
	int		width,
	int		height,
	unsigned char	* rgb,		// NULL, new data, or current value
	unsigned char	* alpha		// NULL, new data, or current value
	);

mtFont * mtimg_font_new (		// Create new font structure, defaults
					// to 12 point size
	char	const	* filename,	// Font file for FreeType to open.
					// NULL = Fixed
	int		face_index	// Usually 0, but maybe higher for
					// bitmap fonts like *.FON.  Fixed fonts
					// are 8 or 16 for the height
	);

int mtimg_font_destroy (		// Destroy and free a font
	mtFont		* font
	);

int mtimg_font_get_height (		// Get font height
	mtFont		* font
	);
	// 0 = Error else height

int mtimg_font_get_width (		// Get font width estimate
	mtFont		* font
	);
	// 0 = Error else width

int mtimg_font_set_size (		// Set font size
	mtFont		* font,
	int		size
	);

mtImage * mtimg_font_render (		// Create image from text
	mtFont		* font,
	char	const	* utf8,
	int		max_width	// 0=No max width
	);



#endif		// MTIMG_H_

